@@ -6,15 +6,30 @@ import types
 import ssl
 import sys
 import urllib3
+import http
 import urllib3.util as urllib_util
+from urllib3.exceptions import HTTPError
 import xml.etree.ElementTree as etree
+try:
+    # Python 3
+    import http.client as httplib
+except ImportError:
+    # Python 2
+    import httplib
 
 from collections import OrderedDict
 from xml.sax.saxutils import escape
-
+import time
+from urllib3.exceptions import ProtocolError, MaxRetryError, ReadTimeoutError
 
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
+def get_http_status_message(status_code):
+    try:
+        return httplib.responses[status_code]
+    except KeyError:
+        return "Unknown Status Code"
+
 if sys.version_info.major == 2:
     unicode = unicode
     str = str
@@ -169,6 +184,20 @@ class CommonResponseInfo:
                 attr_type_name == 'float' or attr_type_name == 'bool' or attr_type_name == 'str')
 
     def _is_empty_object(self, atrrib_data):
+        for key, value in atrrib_data.items():
+            if isinstance(value, list) and len(value) > 0:
+                return False
+
+            if not isinstance(value, list) and value is not None:
+                return False
+
+        return True
+
+    # This is actually the original version of _is_empty_object.
+    # Since it does not consider an empty list as empty (returns False) it caused bugs with setting the
+    # is_empty_object flag, but fixing that caused an issue on populating the data.
+    # So I was forced to preserve the original version for compatibility with the data population logic.
+    def _is_empty_object_except_empty_list(self, atrrib_data):
         for key, value in atrrib_data.items():
             if isinstance(value, list) and len(value) > 0:
                 return False
@@ -189,13 +218,23 @@ class CommonResponseInfo:
                 data = attr_type_instance(list_node, find_prefix)
 
             if not (hasattr(list_node, "attrib") and list_node.attrib):
-                setattr(data, "is_empty_object", True)
+                self._set_is_empty_to_true_if_not_set(data)
 
-        if hasattr(data, "is_empty_object") and data.is_empty_object:
+        if self._is_empty_true(data):
             return None
         else:
             return data
 
+    def _set_is_empty_to_true_if_not_set(self, data):
+        if not hasattr(data, "is_empty_object"):
+            setattr(data, "is_empty_object", True)
+
+    def _set_is_empty_to_false(self, data):
+        setattr(data, "is_empty_object", False)
+        
+    def _is_empty_true(self, data):
+        return hasattr(data, "is_empty_object") and data.is_empty_object
+    
     def _parseAttributesData(self, class_type, xml_object, find_prefix):
         attrib_data_dict = dict()
 
@@ -248,6 +287,10 @@ class CommonResponseInfo:
                             child_count += 1
 
                     # I think that it is a logical bug, but ...
+                    # DannyK: In my opinion, this is done for backwards compatibility with element names that
+                    #         originally represented a single item and later on changed to a list.
+                    #         If a bug related to ghost objects is discovered in the future,
+                    #         this is the first place I would start debugging.        
                     if child_count == 0:
                         for list_node in xml_object:
                             if XMLWrapper.getNodeName(list_node) == name:
@@ -260,10 +303,13 @@ class CommonResponseInfo:
                 attrib_data_dict[name] = data_list
 
         if not self._is_empty_object(attrib_data_dict):
+            self._set_is_empty_to_false(self)
+        elif len(self.__dict__) == empty_object_size:
+            self._set_is_empty_to_true_if_not_set(self)
+
+        if not self._is_empty_object_except_empty_list(attrib_data_dict):
             for key, value in attrib_data_dict.items():
                 setattr(self, key, value)
-        elif len(self.__dict__) == empty_object_size:
-            setattr(self, "is_empty_object", True)
 
 
 class CommonApiResult:
@@ -358,21 +404,33 @@ class CommonAPISession:
         except:
             return unicode(data.decode("utf-8"))
 
-    def _sendRequest(self, operation, message, request_headers):
+    def _sendRequest(self, operation, message, request_headers, attempts=1):
         """ Sending http POST request through URLLIB package
 
         :param operation: operation name
         :param message: request body
         :param request_headers: header of the request
 
-        :return: responce string data
+        :return: http response
         """
         operation_url = str(self.url + operation)
-        response = self._http.request("POST", operation_url, body=message.encode('utf-8'), headers=request_headers)
+        response = retry(self._http.request, ProtocolError, attempts, 1,
+                         method="POST", url=operation_url, body=message.encode('utf-8'), headers=request_headers)
+        
+        self._check_http_response(response)
+        
+        return response
+
+    def _check_http_response(self, response):
+        if not (200 <= response.status < 300):
+            try:
+                status_description = get_http_status_message(response.status)
+            except ValueError:
+                status_description = 'Unknown'
+            raise Exception("({}) [{}] reason:{}".format(status_description, response.status, response.reason))
+        
         if not response.data:
             raise Exception(response.reason)
-        return response.data
-
 
     def _new_serializeRequestData(self, root_node, object_data, prev_type=None):
         """Generate xml from received request data using etree.xml
@@ -432,21 +490,26 @@ class CommonAPISession:
         method_name = kwargs.pop('method_name', None)
 
         message = self._serialize_request(kwargs, method_name)
-
-        result = self._handle_api_response(self._sendRequest(method_name, message))
+        http_response = self._sendRequest(method_name, message)
+        result = self._handle_api_response(http_response)
 
         return result
 
-    def _handle_api_response(self, response_str):
+    def _handle_api_response(self, response):
         '''
-        :param str response_str:
+        :param http response:
         :return: either string or deserialized instance representing the api response
         '''
+        response_str = response.data
         response_str = self._remove_xml_namespace(response_str)
         try:
             api_response = self._deserialize_response(response_str)
         except Exception:
-            raise Exception("Invalid Request: " + str(response_str))
+            try:
+                status_description = get_http_status_message(response.status)
+            except ValueError:
+                status_description = 'Unknown'
+            raise Exception("{} ({}) [{}] reason:{} data:{}".format(response.reason, status_description, response.status, response.reason, str(response_str)))
         if not api_response.success:
             raise CloudShellAPIError(api_response.error_code, api_response.error, response_str)
         result = response_str
@@ -479,3 +542,15 @@ class CommonAPISession:
         rough_string = etree.tostring(elem, 'utf-8')
         reparsed = parseString(rough_string)
         return reparsed.toprettyxml(indent="\t")
+
+def retry(func, exception, attempts=1, delay=1, *args, **kwargs):
+    for i in range(attempts):
+        try:
+            return func(*args, **kwargs)
+        except (MaxRetryError, ProtocolError, ReadTimeoutError, HTTPError) as e:
+            if i == attempts -1:
+                raise e
+            time.sleep(delay)
+            delay *= 2
+        except Exception as e:
+            raise e
